import * as utility from "./toolUtil.js";

//------- APIs to calculate min, max limits and values for a parameter ------//
let DGAIN;
export function calculate_AVDD(args) {
  let ASUPPLY_MIN =
    args.parameterConfigData[args.selectedDevice].ASUPPLY_MIN != undefined
      ? parseFloat(args.parameterConfigData[args.selectedDevice].ASUPPLY_MIN)
      : this.get_ASUPPLY_MIN();
  let ASUPPLY_MAX =
    args.parameterConfigData[args.selectedDevice].ASUPPLY_MAX != undefined
      ? parseFloat(args.parameterConfigData[args.selectedDevice].ASUPPLY_MAX)
      : this.get_ASUPPLY_MAX();
  let AvddVal =
    args.parameterConfigData[args.selectedDevice].AVDD != undefined
      ? args.parameterConfigData[args.selectedDevice].AVDD
      : args.avss.value + ASUPPLY_MAX;
  let val = args.avdd.value != undefined ? args.avdd.value : AvddVal;

  if (ASUPPLY_MIN == ASUPPLY_MAX) {
    return {
      minValue: args.avss.value + ASUPPLY_MIN,
      maxValue: args.avss.value + ASUPPLY_MAX,
      value: val,
      isDisabled: ASUPPLY_MIN == ASUPPLY_MAX,
      absMinValue: val,
      absMaxValue: val,
    };
  }

  return {
    minValue: args.avss.value + ASUPPLY_MIN,
    maxValue: args.avss.value + ASUPPLY_MAX,
    value: val,
    isDisabled: ASUPPLY_MIN == ASUPPLY_MAX,
  };
}

export function calculate_AVSS(args) {
  let avsslabel =
    args.parameterConfigData[args.selectedDevice].AVSS_LABEL != undefined
      ? args.parameterConfigData[args.selectedDevice].AVSS_LABEL
      : args.avss.label;
  let AVSS_MIN =
    args.parameterConfigData[args.selectedDevice].AVSS_MIN != undefined
      ? parseFloat(args.parameterConfigData[args.selectedDevice].AVSS_MIN)
      : this.get_AVSS_MIN();
  let AVSS_MAX =
    args.parameterConfigData[args.selectedDevice].AVSS_MAX != undefined
      ? parseFloat(args.parameterConfigData[args.selectedDevice].AVSS_MAX)
      : this.get_AVSS_MAX();
  if (AVSS_MIN == AVSS_MAX) {
    return {
      label: avsslabel,
      minValue: AVSS_MIN,
      maxValue: AVSS_MAX,
      isDisabled: AVSS_MIN == AVSS_MAX,
      absMinValue: AVSS_MAX,
      absMaxValue: AVSS_MIN,
    };
  }

  return {
    label: avsslabel,
    minValue: AVSS_MIN,
    maxValue: AVSS_MAX,
    isDisabled: AVSS_MIN == AVSS_MAX,
  };
}

export function calculate_PGA_GAIN(args) {
  let configData = args.parameterConfigData[args.selectedDevice];
  if (configData && configData.pga_gain) {
    let internalGainObject = get_Control_Object(configData.pga_gain);
    DGAIN = args.parameterConfigData[args.selectedDevice].DGAIN
      ? typeof args.parameterConfigData[args.selectedDevice].DGAIN === "object"
        ? args.parameterConfigData[args.selectedDevice].DGAIN[
            args.pga_gain.value != undefined
              ? args.pga_gain.value
              : internalGainObject.value
          ]
        : parseFloat(args.parameterConfigData[args.selectedDevice].DGAIN)
      : this.get_DGAIN(
          args.pga_gain.value != undefined
            ? args.pga_gain.value
            : internalGainObject.value,
        );
    return {
      options: internalGainObject.options,
      value:
        args.pga_gain.value != undefined
          ? args.pga_gain.value
          : internalGainObject.value,
    };
  }
  return { options: this.get_GAIN_LIST() };
}

export function calculate_VREFN(args) {
  let pga_gain = this.compute_pga_gain(args);
  let VREF_DIFF_MIN =
    args.parameterConfigData[args.selectedDevice].VREF_DIFF_MIN != undefined
      ? parseFloat(args.parameterConfigData[args.selectedDevice].VREF_DIFF_MIN)
      : this.get_VREF_DIFF_MIN();

  let VREFN_TO_AVSS_MAX =
    args.parameterConfigData[args.selectedDevice].VREFN_TO_AVSS_MAX != undefined
      ? parseFloat(
          args.parameterConfigData[args.selectedDevice].VREFN_TO_AVSS_MAX,
        )
      : this.get_VREFN_TO_AVSS_MAX(pga_gain);

  let VREFN_DISABLED =
    args.parameterConfigData[args.selectedDevice].VREFN_DISABLED != undefined
      ? parseFloat(args.parameterConfigData[args.selectedDevice].VREFN_DISABLED)
      : undefined;

  return VREFN_DISABLED != undefined
    ? {
        absMinValue: VREFN_DISABLED,
        absMaxValue: VREFN_DISABLED,
        minValue: VREFN_DISABLED,
        maxValue: VREFN_DISABLED,
        value: VREFN_DISABLED,
        isDisabled: true,
      }
    : {
        absMinValue: args.avss.value - 0.3,
        absMaxValue: args.avdd.value + 0.3,
        minValue: args.avss.value - VREFN_TO_AVSS_MAX,
        maxValue: args.vrefp.value - VREF_DIFF_MIN,
        isDisabled: false,
      };
}

export function calculate_VREFP(args) {
  let configData = args.parameterConfigData[args.selectedDevice];
  let refSource;
  if (
    configData.controls != undefined &&
    configData.controls.configuration.refSource != undefined
  ) {
    refSource = configData.controls.configuration.refSource.value;
  }
  if (refSource != undefined && configData.vrefpControl != undefined) {
    return {
      absMinValue: configData.vrefpControl[refSource].VREF_DIFF_MIN,
      absMaxValue: configData.vrefpControl[refSource].VREF_DIFF_MAX,
      isDisabled:
        configData.vrefpControl[refSource].VREF_DIFF_MIN ==
        configData.vrefpControl[refSource].VREF_DIFF_MAX,
    };
  }
  let pga_gain = this.compute_pga_gain(args);
  let VREF_DIFF_MIN =
    args.parameterConfigData[args.selectedDevice].VREF_DIFF_MIN != undefined
      ? parseFloat(args.parameterConfigData[args.selectedDevice].VREF_DIFF_MIN)
      : this.get_VREF_DIFF_MIN();
  let VREF_DIFF_MAX =
    args.parameterConfigData[args.selectedDevice].VREF_DIFF_MAX != undefined
      ? parseFloat(args.parameterConfigData[args.selectedDevice].VREF_DIFF_MAX)
      : this.get_VREF_DIFF_MAX();
  if (VREF_DIFF_MIN == VREF_DIFF_MAX) {
    return {
      absMinValue: VREF_DIFF_MIN,
      absMaxValue: VREF_DIFF_MAX,
      value: VREF_DIFF_MIN,
      isDisabled: true,
    };
  }

  let VREFP_TO_AVDD_MAX =
    args.parameterConfigData[args.selectedDevice].VREFP_TO_AVDD_MAX != undefined
      ? parseFloat(
          args.parameterConfigData[args.selectedDevice].VREFP_TO_AVDD_MAX,
        )
      : this.get_VREFP_TO_AVDD_MAX(pga_gain);
  return {
    absMinValue: args.avss.value - 0.3,
    absMaxValue: args.avdd.value + 0.3,
    minValue: args.vrefn.value + VREF_DIFF_MIN,
    maxValue: Math.min(
      args.avdd.value + VREFP_TO_AVDD_MAX,
      args.vrefn.value +
        this.get_VREF_DIFF_MAX(args.avdd.value, args.avss.value),
    ),
    isDisabled: false,
  };
}

export function calculate_AINP(args) {
  let configData = args.parameterConfigData[args.selectedDevice];
  let pga_gain = this.compute_pga_gain(args);
  if (args.input_mode.value === "Differential/CM Voltages") {
    return {
      absMinValue: "",
      absMaxValue: "",
      isDisabled: true,
      minValue: undefined,
      maxValue: undefined,
      removeTooltip: true,
      value: args.vin_cm.value + args.vin_diff.value / 2,
    };
  }

  const VIN_DIFF_MAX = this.get_VIN_DIFF_MAX(
    args.avdd.value,
    args.avss.value,
    pga_gain,
    args.vrefn.value,
    args.vrefp.value,
    args,
  );

  // min value dependencies
  const AINx_ABS_MIN = this.compute_AINx_ABS_MIN(args);

  let AinpVal =
    args.parameterConfigData[args.selectedDevice].AINP != undefined
      ? args.parameterConfigData[args.selectedDevice].AINP
      : args.ainp.value;
  let val = args.ainp.value != undefined ? args.ainp.value : AinpVal;

  const AINP_DIFF_NEG_CLIP = this.getValueForClip_AINP(
    "neg",
    args.ainn.value,
    VIN_DIFF_MAX,
  );
  const AINP_VOUTP_NEG_SAT = this.get_AINP_VOUTP_NEG_SAT(
    args.ainn.value,
    args.avss.value,
    pga_gain,
    args,
  );
  const AINP_VOUTN_POS_SAT = this.get_AINP_VOUTN_POS_SAT(
    args.ainn.value,
    args.avdd.value,
    pga_gain,
    args,
  );

  // max value dependencies
  const AINx_ABS_MAX = this.compute_AINx_ABS_MAX(args);

  const AINP_DIFF_POS_CLIP = this.getValueForClip_AINP(
    "pos",
    args.ainn.value,
    VIN_DIFF_MAX,
  );
  const AINP_VOUTP_POS_SAT = this.get_AINP_VOUTP_POS_SAT(
    args.ainn.value,
    args.avdd.value,
    pga_gain,
    args,
  );
  const AINP_VOUTN_NEG_SAT = this.get_AINP_VOUTN_NEG_SAT(
    args.ainn.value,
    args.avss.value,
    pga_gain,
    args,
  );

  return {
    absMinValue: args.avss.value - 0.3,
    absMaxValue: args.avdd.value + 0.3,
    isDisabled: false,
    value: val,
    removeTooltip: false,
    minValue: Math.max(
      ...[
        AINx_ABS_MIN,
        AINP_DIFF_NEG_CLIP,
        AINP_VOUTP_NEG_SAT,
        AINP_VOUTN_POS_SAT,
      ].filter(value => !isNaN(value)),
    ),
    maxValue: Math.min(
      ...[
        AINx_ABS_MAX,
        AINP_DIFF_POS_CLIP,
        AINP_VOUTP_POS_SAT,
        AINP_VOUTN_NEG_SAT,
      ].filter(value => !isNaN(value)),
    ),
  };
}

export function calculate_AINN(args) {
  let configData = args.parameterConfigData[args.selectedDevice];
  let pga_gain = this.compute_pga_gain(args);
  if (args.input_mode.value === "Differential/CM Voltages") {
    return {
      absMinValue: "",
      absMaxValue: "",
      isDisabled: true,
      minValue: undefined,
      maxValue: undefined,
      removeTooltip: true,
      value: args.vin_cm.value - args.vin_diff.value / 2,
    };
  }

  const VIN_DIFF_MAX = this.get_VIN_DIFF_MAX(
    args.avdd.value,
    args.avss.value,
    pga_gain,
    args.vrefn.value,
    args.vrefp.value,
    args,
  );

  // min value dependencies
  const AINx_ABS_MIN = this.compute_AINx_ABS_MIN(args);

  let AinnVal =
    args.parameterConfigData[args.selectedDevice].AINN != undefined
      ? args.parameterConfigData[args.selectedDevice].AINN
      : args.ainn.value;
  let val = args.ainn.value != undefined ? args.ainn.value : AinnVal;

  const AINN_DIFF_POS_CLIP = this.getValueForClip_AINN(
    "pos",
    args.ainp.value,
    VIN_DIFF_MAX,
  );
  const AINN_VOUTN_NEG_SAT = this.get_AINN_VOUTN_NEG_SAT(
    args.ainp.value,
    args.avss.value,
    pga_gain,
    args,
  );
  const AINN_VOUTP_POS_SAT = this.get_AINN_VOUTP_POS_SAT(
    args.ainp.value,
    args.avdd.value,
    pga_gain,
    args,
  );

  // max value dependencies
  const AINx_ABS_MAX = this.compute_AINx_ABS_MAX(args);
  const AINN_DIFF_NEG_CLIP = this.getValueForClip_AINN(
    "neg",
    args.ainp.value,
    VIN_DIFF_MAX,
  );
  const AINN_VOUTN_POS_SAT = this.get_AINN_VOUTN_POS_SAT(
    args.ainp.value,
    args.avdd.value,
    pga_gain,
    args,
  );
  const AINN_VOUTP_NEG_SAT = this.get_AINN_VOUTP_NEG_SAT(
    args.ainp.value,
    args.avss.value,
    pga_gain,
    args,
  );

  return {
    absMinValue: args.avss.value - 0.3,
    absMaxValue: args.avdd.value + 0.3,
    isDisabled: false,
    removeTooltip: false,
    value: val,
    minValue: Math.max(
      ...[
        AINx_ABS_MIN,
        AINN_DIFF_POS_CLIP,
        AINN_VOUTN_NEG_SAT,
        AINN_VOUTP_POS_SAT,
      ].filter(value => !isNaN(value)),
    ),
    maxValue: Math.min(
      ...[
        AINx_ABS_MAX,
        AINN_DIFF_NEG_CLIP,
        AINN_VOUTN_POS_SAT,
        AINN_VOUTP_NEG_SAT,
      ].filter(value => !isNaN(value)),
    ),
  };
}

export function calculate_VIN_DIFF(args) {
  let pga_gain = this.compute_pga_gain(args);
  const VIN_DIFF_MAX = this.get_VIN_DIFF_MAX(
    args.avdd.value,
    args.avss.value,
    pga_gain,
    args.vrefn.value,
    args.vrefp.value,
    args,
  );

  let updatedVinDiffConfig = {
    absMinValue: args.avss.value - 0.3,
    absMaxValue: args.avdd.value + 0.3,
    isDisabled: false,
    minValue: -VIN_DIFF_MAX,
    maxValue: VIN_DIFF_MAX,
    removeTooltip: false,
  };

  if (args.input_mode.value === "Absolute Voltages") {
    updatedVinDiffConfig["absMaxValue"] = "";
    updatedVinDiffConfig["absMinValue"] = "";
    updatedVinDiffConfig["value"] = args.ainp.value - args.ainn.value;
    updatedVinDiffConfig["isDisabled"] = true;
    updatedVinDiffConfig["removeTooltip"] = true;
  }

  return updatedVinDiffConfig;
}

export function calculate_VIN_CM(args) {
  let configData = args.parameterConfigData[args.selectedDevice];
  let pga_gain = this.compute_pga_gain(args);
  let calculated_AGAIN =
    args.parameterConfigData[args.selectedDevice].AGAIN != undefined
      ? typeof args.parameterConfigData[args.selectedDevice].AGAIN === "object"
        ? args.parameterConfigData[args.selectedDevice].AGAIN[pga_gain]
        : parseInt(args.parameterConfigData[args.selectedDevice].AGAIN)
      : this.get_AGAIN(pga_gain);
  const modifiedVinDiff = Math.abs(
    (calculated_AGAIN * args.vin_diff.value) / 2,
  );

  let updatedVinCmConfig = {
    absMinValue: args.avss.value - 0.3,
    absMaxValue: args.avdd.value + 0.3,
    removeTooltip: false,
    isDisabled: false,
    minValue:
      this.get_VOUT_MIN(args.avss.value, pga_gain, args) + modifiedVinDiff,
    maxValue:
      this.get_VOUT_MAX(args.avdd.value, pga_gain, args) - modifiedVinDiff,
  };

  if (args.input_mode.value === "Absolute Voltages") {
    updatedVinCmConfig["absMaxValue"] = "";
    updatedVinCmConfig["absMinValue"] = "";
    updatedVinCmConfig["value"] = (args.ainp.value + args.ainn.value) / 2;
    updatedVinCmConfig["isDisabled"] = true;
    updatedVinCmConfig["removeTooltip"] = true;
  }

  return updatedVinCmConfig;
}

//------- API to form Graph meta data ------//
export function computeCmRangeGraph(args) {
  let pga_gain = this.compute_pga_gain(args);
  const vout_cm_clampped = this.get_VOUT_CM_CLAMPED(
    args.avdd.value,
    args.avss.value,
    pga_gain,
    args.vin_cm.value,
    args.vin_diff.value,
    args,
  );

  const vref_diff = this.get_VREF_DIFF(args.vrefn.value, args.vrefp.value);

  const pga_limits = this.get_PGA_INPUT_LIMITS(
    args.avdd.value,
    args.avss.value,
    pga_gain,
    args.vin_cm.maxValue,
    args.vin_cm.minValue,
    args.vin_diff.value,
    args.vrefn.value,
    args.vrefp.value,
    args,
  );

  return {
    metaData: {
      avdd: args.avdd.value,
      avss: args.avss.value,
      ainp: args.ainp.value,
      ainn: args.ainn.value,
      graph_image:
        args.parameterConfigData[args.selectedDevice].graph_image != undefined
          ? args.parameterConfigData[args.selectedDevice].graph_image
          : args.graph_image.value,

      voutp_clamped: this.get_VOUTP_CLAMPED(
        args.avdd.value,
        args.avss.value,
        pga_gain,
        args.vin_cm.value,
        args.vin_diff.value,
        args,
      ),
      voutn_clamped: this.get_VOUTN_CLAMPED(
        args.avdd.value,
        args.avss.value,
        pga_gain,
        args.vin_cm.value,
        args.vin_diff.value,
        args,
      ),
      vout_max: this.get_VOUT_MAX(args.avdd.value, pga_gain, args),
      vout_min: this.get_VOUT_MIN(args.avss.value, pga_gain, args),

      vref_positive_limit: vout_cm_clampped + vref_diff / (2 * DGAIN),
      vref_negative_limit: vout_cm_clampped - vref_diff / (2 * DGAIN),

      pga_input_pos_limit: pga_limits.posLimit,
      pga_input_neg_limit: pga_limits.negLimit,

      error_voutp: {
        status: this.check_ERROR_VOUTP(
          args.avdd.value,
          args.avss.value,
          pga_gain,
          args.vin_cm.value,
          args.vin_diff.value,
          args,
        ),
        text: "VOUTP is saturated",
      },

      error_voutn: {
        status: this.check_ERROR_VOUTN(
          args.avdd.value,
          args.avss.value,
          pga_gain,
          args.vin_cm.value,
          args.vin_diff.value,
          args,
        ),
        text: "VOUTN is saturated",
      },

      error_vout_diff_exceeds_fsr: {
        status: this.check_ERROR_VOUT_DIFF_EXCEEDS_FSR(
          args.avdd.value,
          args.avss.value,
          pga_gain,
          args.vin_cm.value,
          args.vin_diff.value,
          args.vrefn.value,
          args.vrefp.value,
          args,
        ),
        text: "Differential output voltage too large<br>(output codes clipped)",
      },
    },
  };
}

// ---------- CM calculator's other parameter calculations ---------------//
export function get_BUFFER_ENABLED(GAIN) {
  if (GAIN === "Bypassed") return false;
  return true;
}

export function get_GAIN_LIST() {
  return ["Bypassed", "1", "2", "4", "8", "16", "32", "64", "128"];
}

export function get_AGAIN(GAIN) {
  const AnalogGainForPgaGain = {
    Bypassed: 1,
    1: 1,
    2: 2,
    4: 4,
    8: 8,
    16: 16,
    32: 32,
    64: 32,
    128: 32,
  };

  return AnalogGainForPgaGain[GAIN];
}

export function get_DGAIN(GAIN) {
  const DigitalGainForPgaGain = {
    Bypassed: 1,
    1: 1,
    2: 1,
    4: 1,
    8: 1,
    16: 1,
    32: 1,
    64: 2,
    128: 4,
  };

  return DigitalGainForPgaGain[GAIN];
}

export function getEquivalentGain(GAIN, args) {
  return args.pga_gain.options[GAIN] === "Bypassed"
    ? 1
    : parseInt(args.pga_gain.options[GAIN]);
}

export function get_VREF_FSR_SF() {
  return 2;
}

export function get_POS_SWING_TO_RAIL(isBufferEnabled) {
  if (!isBufferEnabled) return -0.05;
  else return 0.15;
}

export function get_NEG_SWING_TO_RAIL(isBufferEnabled) {
  if (!isBufferEnabled) return -0.05;
  else return 0.15;
}

export function get_AINx_ABS_MIN(AVSS) {
  return AVSS - 0.3;
}

export function get_AINx_ABS_MAX(AVDD) {
  return AVDD + 0.3;
}

export function get_AVSS_MIN() {
  return -2.625;
}

export function get_AVSS_MAX() {
  return 0.05;
}

export function get_ASUPPLY_MIN() {
  return 2.7;
}

export function get_ASUPPLY_MAX() {
  return 5.25;
}

export function get_VREF_DIFF_MIN() {
  return 0.5;
}

export function get_VREF_DIFF_MAX(AVDD, AVSS) {
  return this.get_ASUPPLY(AVDD, AVSS);
}

export function get_VREFP_TO_AVDD_MAX(GAIN) {
  if (!this.get_BUFFER_ENABLED(GAIN)) return 0.05;
  else return 0;
}

export function get_VREFN_TO_AVSS_MAX(GAIN) {
  if (!this.get_BUFFER_ENABLED(GAIN)) return 0.05;
  else return 0;
}

// ----------- Other background calculaions ---------- //
export function get_VREF_DIFF(VREFN, VREFP) {
  return Math.abs(VREFP - VREFN);
}

export function get_FSR(GAIN, VREFN, VREFP, args) {
  let configData = args.parameterConfigData[args.selectedDevice];
  if (configData && configData.VREF_FSR_SF) {
    let VREF_FSR_SF = parseFloat(configData.VREF_FSR_SF);
    if (configData.controls != undefined) {
      VREF_FSR_SF = configData.controls.configuration.refSource
        ? configData.VREF_FSR_SF[
            configData.controls.configuration.refSource.value
          ]
        : parseFloat(configData.VREF_FSR_SF);
    }

    return Math.abs((this.get_VREF_DIFF(VREFN, VREFP) * VREF_FSR_SF) / GAIN);
  }
  return Math.abs(
    (this.get_VREF_DIFF(VREFN, VREFP) * this.get_VREF_FSR_SF()) /
      this.getEquivalentGain(GAIN, args),
  );
}

export function get_ASUPPLY(AVDD, AVSS) {
  return Math.abs(AVDD - AVSS);
}

export function get_PGA_OUTPUT_SWING_MAX(ASUPPLY, GAIN, args) {
  const BUFFER_ENABLED = this.get_BUFFER_ENABLED(GAIN);
  let configData = args.parameterConfigData[args.selectedDevice];

  if (
    configData &&
    configData.POS_SWING_TO_RAIL &&
    configData.NEG_SWING_TO_RAIL
  ) {
    if (configData.POS_SWING_TO_RAIL.bufferEnabled == undefined) {
      return Math.abs(
        ASUPPLY -
          (parseFloat(configData.POS_SWING_TO_RAIL[GAIN]) +
            parseFloat(configData.NEG_SWING_TO_RAIL[GAIN])),
      );
    }

    return Math.abs(
      ASUPPLY -
        (BUFFER_ENABLED
          ? parseFloat(configData.POS_SWING_TO_RAIL.bufferEnabled)
          : parseFloat(configData.POS_SWING_TO_RAIL.bufferDisabled) +
            BUFFER_ENABLED
          ? parseFloat(configData.NEG_SWING_TO_RAIL.bufferEnabled)
          : parseFloat(configData.NEG_SWING_TO_RAIL.bufferDisabled)),
    );
  }
  return Math.abs(
    ASUPPLY -
      (this.get_POS_SWING_TO_RAIL(BUFFER_ENABLED) +
        this.get_NEG_SWING_TO_RAIL(BUFFER_ENABLED)),
  );
}

export function get_VIN_DIFF_MAX(AVDD, AVSS, GAIN, VREFN, VREFP, args) {
  return Math.abs(VREFP - VREFN) / GAIN;
}

export function get_VOUTP_IDEAL(GAIN, VIN_CM, VIN_DIFF, args) {
  let calculated_AGAIN =
    args.parameterConfigData[args.selectedDevice].AGAIN != undefined
      ? typeof args.parameterConfigData[args.selectedDevice].AGAIN === "object"
        ? args.parameterConfigData[args.selectedDevice].AGAIN[GAIN]
        : parseInt(args.parameterConfigData[args.selectedDevice].AGAIN)
      : this.get_AGAIN(GAIN);
  return VIN_CM + (calculated_AGAIN * VIN_DIFF) / 2;
}

export function get_VOUTN_IDEAL(GAIN, VIN_CM, VIN_DIFF, args) {
  let calculated_AGAIN =
    args.parameterConfigData[args.selectedDevice].AGAIN != undefined
      ? typeof args.parameterConfigData[args.selectedDevice].AGAIN === "object"
        ? args.parameterConfigData[args.selectedDevice].AGAIN[GAIN]
        : parseInt(args.parameterConfigData[args.selectedDevice].AGAIN)
      : this.get_AGAIN(GAIN);
  return VIN_CM - (calculated_AGAIN * VIN_DIFF) / 2;
}

export function get_VOUT_DIFF_IDEAL(GAIN, VIN_CM, VIN_DIFF, args) {
  return (
    this.get_VOUTP_IDEAL(GAIN, VIN_CM, VIN_DIFF, args) -
    this.get_VOUTN_IDEAL(GAIN, VIN_CM, VIN_DIFF, args)
  );
}

export function get_VOUT_MAX(AVDD, GAIN, args) {
  let configData = args.parameterConfigData[args.selectedDevice];
  if (configData && configData.POS_SWING_TO_RAIL != undefined) {
    if (configData.POS_SWING_TO_RAIL.bufferEnabled != undefined) {
      return args.pga_gain.value
        ? AVDD - parseFloat(configData.POS_SWING_TO_RAIL.bufferEnabled)
        : AVDD - parseFloat(configData.POS_SWING_TO_RAIL.bufferDisabled);
    }
    if (configData.POS_SWING_TO_RAIL.bufferDisabled == undefined) {
      return AVDD - parseFloat(configData.POS_SWING_TO_RAIL[GAIN]);
    }
  }
  return (
    AVDD - parseFloat(this.get_POS_SWING_TO_RAIL(this.get_BUFFER_ENABLED(GAIN)))
  );
  // return AVDD - this.get_POS_SWING_TO_RAIL(this.get_BUFFER_ENABLED(GAIN));
}

export function get_VOUT_MIN(AVSS, GAIN, args) {
  let configData = args.parameterConfigData[args.selectedDevice];
  if (configData && configData.NEG_SWING_TO_RAIL != undefined) {
    if (configData.NEG_SWING_TO_RAIL.bufferEnabled != undefined) {
      return args.pga_gain.value
        ? AVSS + parseFloat(configData.NEG_SWING_TO_RAIL.bufferEnabled)
        : AVSS + parseFloat(configData.NEG_SWING_TO_RAIL.bufferDisabled);
    }
    if (configData.NEG_SWING_TO_RAIL.bufferDisabled == undefined) {
      return AVSS + parseFloat(configData.NEG_SWING_TO_RAIL[GAIN]);
    }
  }
  return (
    AVSS + parseFloat(this.get_NEG_SWING_TO_RAIL(this.get_BUFFER_ENABLED(GAIN)))
  );
  // return AVSS + this.get_NEG_SWING_TO_RAIL(this.get_BUFFER_ENABLED(GAIN));
}

export function get_VOUTP_CLAMPED(AVDD, AVSS, GAIN, VIN_CM, VIN_DIFF, args) {
  const VOUT_MAX = this.get_VOUT_MAX(AVDD, GAIN, args);
  const VOUT_MIN = this.get_VOUT_MIN(AVSS, GAIN, args);
  const VOUTP_IDEAL = this.get_VOUTP_IDEAL(GAIN, VIN_CM, VIN_DIFF, args);

  if (VOUTP_IDEAL > VOUT_MAX) return VOUT_MAX;
  if (VOUTP_IDEAL < VOUT_MIN) return VOUT_MIN;
  return VOUTP_IDEAL;
}

export function get_VOUTN_CLAMPED(AVDD, AVSS, GAIN, VIN_CM, VIN_DIFF, args) {
  const VOUT_MAX = this.get_VOUT_MAX(AVDD, GAIN, args);
  const VOUT_MIN = this.get_VOUT_MIN(AVSS, GAIN, args);
  const VOUTN_IDEAL = this.get_VOUTN_IDEAL(GAIN, VIN_CM, VIN_DIFF, args);
  if (VOUTN_IDEAL > VOUT_MAX) return VOUT_MAX;
  if (VOUTN_IDEAL < VOUT_MIN) return VOUT_MIN;
  return VOUTN_IDEAL;
}

export function get_VOUT_DIFF_CLAMPED(
  AVDD,
  AVSS,
  GAIN,
  VIN_CM,
  VIN_DIFF,
  args,
) {
  return (
    this.get_VOUTP_CLAMPED(AVDD, AVSS, GAIN, VIN_CM, VIN_DIFF, args) -
    this.get_VOUTN_CLAMPED(AVDD, AVSS, GAIN, VIN_CM, VIN_DIFF, args)
  );
}

export function get_VOUT_CM_CLAMPED(AVDD, AVSS, GAIN, VIN_CM, VIN_DIFF, args) {
  return (
    (this.get_VOUTP_CLAMPED(AVDD, AVSS, GAIN, VIN_CM, VIN_DIFF, args) +
      this.get_VOUTN_CLAMPED(AVDD, AVSS, GAIN, VIN_CM, VIN_DIFF, args)) /
    2
  );
}

export function compute_AINx_ABS_MIN(args) {
  let minVal =
    args.parameterConfigData[args.selectedDevice].AINx_ABS_MIN != undefined
      ? parseFloat(
          args.avss.value -
            args.parameterConfigData[args.selectedDevice].AINx_ABS_MIN,
        )
      : this.get_AINx_ABS_MIN(args.avss.value);
  return minVal;
}
export function compute_AINx_ABS_MAX(args) {
  let maxVal =
    args.parameterConfigData[args.selectedDevice].AINx_ABS_MAX != undefined
      ? parseFloat(
          args.avdd.value +
            args.parameterConfigData[args.selectedDevice].AINx_ABS_MAX,
        )
      : this.get_AINx_ABS_MAX(args.avdd.value);
  return maxVal;
}

export function compute_pga_gain(args) {
  let pga_gain =
    args.pga_gain.value != undefined
      ? this.getEquivalentGain(args.pga_gain.value, args)
      : this.getEquivalentGain(
          get_Control_Object(
            args.parameterConfigData[args.selectedDevice].pga_gain,
          ).value,
        );
  return pga_gain;
}

export function get_PGA_INPUT_LIMITS(
  AVDD,
  AVSS,
  GAIN,
  VCM_MAX,
  VCM_MIN,
  VIN_DIFF,
  VREFN,
  VREFP,
  args,
) {
  const VIN_DIFF_MAX = this.get_VIN_DIFF_MAX(
    AVDD,
    AVSS,
    GAIN,
    VREFN,
    VREFP,
    args,
  );

  if (Math.abs(VIN_DIFF) > VIN_DIFF_MAX) {
    VCM_MAX = this.get_ASUPPLY(AVDD, AVSS) / 2;
    VCM_MIN = VCM_MAX;
  }

  return {
    posLimit: Math.min(
      VCM_MAX + VIN_DIFF_MAX / 2,
      this.get_VOUT_MAX(AVDD, GAIN, args),
    ),
    negLimit: Math.max(
      VCM_MIN - VIN_DIFF_MAX / 2,
      this.get_VOUT_MIN(AVSS, GAIN, args),
    ),
  };
}

// ------------ Voltage limits for analog inputs ---------//
export function getValueForClip_AINP(condition, AINN, VIN_DIFF_MAX) {
  switch (condition) {
    case "pos":
      return AINN + VIN_DIFF_MAX;
    case "neg":
      return AINN - VIN_DIFF_MAX;
    default:
      return undefined;
  }
}

export function getValueForClip_AINN(condition, AINP, VIN_DIFF_MAX) {
  switch (condition) {
    case "pos":
      return AINP - VIN_DIFF_MAX;
    case "neg":
      return AINP + VIN_DIFF_MAX;
    default:
      return undefined;
  }
}

export function g(GAIN, x, y, args) {
  const AGAIN = this.get_AGAIN(GAIN);
  let calculated_AGAIN =
    args.parameterConfigData[args.selectedDevice].AGAIN != undefined
      ? typeof args.parameterConfigData[args.selectedDevice].AGAIN === "object"
        ? args.parameterConfigData[args.selectedDevice].AGAIN[GAIN]
        : parseInt(args.parameterConfigData[args.selectedDevice].AGAIN)
      : AGAIN;
  return (2 * y - x * (1 - calculated_AGAIN)) / (1 + calculated_AGAIN);
}

export function h(GAIN, x, y, args) {
  const AGAIN = this.get_AGAIN(GAIN);
  let calculated_AGAIN =
    args.parameterConfigData[args.selectedDevice].AGAIN != undefined
      ? typeof args.parameterConfigData[args.selectedDevice].AGAIN === "object"
        ? args.parameterConfigData[args.selectedDevice].AGAIN[GAIN]
        : parseInt(args.parameterConfigData[args.selectedDevice].AGAIN)
      : AGAIN;

  if (calculated_AGAIN === 1) return NaN;

  return (2 * y - x * (1 + calculated_AGAIN)) / (1 - calculated_AGAIN);
}

export function get_AINP_VOUTP_POS_SAT(AINN, AVDD, GAIN, args) {
  return this.g(GAIN, AINN, this.get_VOUT_MAX(AVDD, GAIN, args), args);
}

export function get_AINP_VOUTP_NEG_SAT(AINN, AVSS, GAIN, args) {
  return this.g(GAIN, AINN, this.get_VOUT_MIN(AVSS, GAIN, args), args);
}

export function get_AINP_VOUTN_POS_SAT(AINN, AVDD, GAIN, args) {
  return this.h(GAIN, AINN, this.get_VOUT_MAX(AVDD, GAIN, args), args);
}

export function get_AINP_VOUTN_NEG_SAT(AINN, AVSS, GAIN, args) {
  return this.h(GAIN, AINN, this.get_VOUT_MIN(AVSS, GAIN, args), args);
}

export function get_AINN_VOUTP_POS_SAT(AINP, AVDD, GAIN, args) {
  return this.h(GAIN, AINP, this.get_VOUT_MAX(AVDD, GAIN, args), args);
}

export function get_AINN_VOUTP_NEG_SAT(AINP, AVSS, GAIN, args) {
  return this.h(GAIN, AINP, this.get_VOUT_MIN(AVSS, GAIN, args), args);
}

export function get_AINN_VOUTN_POS_SAT(AINP, AVDD, GAIN, args) {
  return this.g(GAIN, AINP, this.get_VOUT_MAX(AVDD, GAIN, args), args);
}

export function get_AINN_VOUTN_NEG_SAT(AINP, AVSS, GAIN, args) {
  return this.g(GAIN, AINP, this.get_VOUT_MIN(AVSS, GAIN, args), args);
}

// ---------- Error indicators ---------- //
export function check_ERROR_VOUTP(AVDD, AVSS, GAIN, VIN_CM, VIN_DIFF, args) {
  return (
    this.get_VOUTP_IDEAL(GAIN, VIN_CM, VIN_DIFF, args) !==
    this.get_VOUTP_CLAMPED(AVDD, AVSS, GAIN, VIN_CM, VIN_DIFF, args)
  );
}

export function check_ERROR_VOUTN(AVDD, AVSS, GAIN, VIN_CM, VIN_DIFF, args) {
  return (
    this.get_VOUTN_IDEAL(GAIN, VIN_CM, VIN_DIFF, args) !==
    this.get_VOUTN_CLAMPED(AVDD, AVSS, GAIN, VIN_CM, VIN_DIFF, args)
  );
}

export function check_ERROR_VOUT_DIFF_EXCEEDS_FSR(
  AVDD,
  AVSS,
  GAIN,
  VIN_CM,
  VIN_DIFF,
  VREFN,
  VREFP,
  args,
) {
  let configData = args.parameterConfigData[args.selectedDevice];
  if (configData && configData.VREF_FSR_SF) {
    let VREF_FSR_SF = parseFloat(configData.VREF_FSR_SF);
    if (configData.controls != undefined) {
      VREF_FSR_SF = configData.controls.configuration.refSource
        ? configData.VREF_FSR_SF[
            configData.controls.configuration.refSource.value
          ]
        : parseFloat(configData.VREF_FSR_SF);
    }
    return (
      Math.abs(
        this.get_VOUT_DIFF_CLAMPED(AVDD, AVSS, GAIN, VIN_CM, VIN_DIFF, args),
      ) >
      (this.get_VREF_DIFF(VREFN, VREFP) * VREF_FSR_SF) / (2 * DGAIN)
    );
  }
  return (
    Math.abs(
      this.get_VOUT_DIFF_CLAMPED(AVDD, AVSS, GAIN, VIN_CM, VIN_DIFF, args),
    ) >
    (this.get_VREF_DIFF(VREFN, VREFP) * this.get_VREF_FSR_SF()) /
      (2 * this.get_DGAIN(GAIN))
  );
}

function get_Control_Object(controlData) {
  if (typeof controlData === "string" && controlData.includes("|")) {
    // Split the string by pipe characters (|)
    const valueArray = controlData.split("|");

    // Find the index of the element containing "(default)"
    const defaultValueIndex = valueArray.findIndex(element =>
      element.includes("(default)"),
    );

    // Map the array values to an object with indices
    const controlObject = {
      options: valueArray.reduce((acc, value, index) => {
        acc[index] = value.replace("(default)", "").trim();
        return acc;
      }, {}),
      value: defaultValueIndex !== -1 ? defaultValueIndex : 0,
    };

    return controlObject;
  }
  // Return an empty object if the input is not a string or object
  return {};
}
